unit Unit1;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls, Vcl.Samples.Spin,
  VirtualTrees, Vcl.ComCtrls, Vcl.ExtCtrls, Vcl.Samples.Gauges;

type
  TMyStreamReader = class
  private
//    FFileStream: TFileStream;
    FStreamReader: TStreamReader;
    FLineByteCount: Integer;
    FLineCount: Integer;
    function GetLineCount: Integer;
    function GetEndOfStream: boolean;
  public
    constructor Create(const aFileName: TFileName);
    destructor Destroy; override;
    function ReadLine(): string; overload;
    function ReadLine(const aLine: integer): string; overload;
    property LineCount: Integer read GetLineCount;
    property EndOfStream: boolean read GetEndOfStream;
  end;

  TForm1 = class(TForm)
    btnCreateLargeFile: TButton;
    VirtualStringTree1: TVirtualStringTree;
    btnFillVST: TButton;
    btnClearVST: TButton;
    TreeView1: TTreeView;
    btnClearTV: TButton;
    btnFillTV: TButton;
    grpVST: TGroupBox;
    grpTV: TGroupBox;
    lblFillControlElapsed: TLabel;
    lblMemory: TLabel;
    timMemoryTrace: TTimer;
    grpStatistics: TGroupBox;
    gauMemory: TGauge;
    speFileLinesCount: TSpinEdit;
    lblFileLinesCount: TLabel;
    lblVSTLinesCount: TLabel;
    lblTVLinesCount: TLabel;
    lblMemoryMax: TLabel;
    procedure btnCreateLargeFileClick(Sender: TObject);
    procedure VirtualStringTree1GetText(Sender: TBaseVirtualTree;
      Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType;
      var CellText: string);
    procedure btnFillVSTClick(Sender: TObject);
    procedure btnClearVSTClick(Sender: TObject);
    procedure btnFillTVClick(Sender: TObject);
    procedure btnClearTVClick(Sender: TObject);
    procedure timMemoryTraceTimer(Sender: TObject);
  private
  const
    cLargeFile: string = '../../TextFile#2.txt';
      {$J+}
//    cFillControlElapsed: string = '';
//    {$J-}
  private
    { Private declarations }
    FFillControlElapsedCaption: string;
    FVSTLinesCountCaption: string;
    FTVLinesCountCaption: string;
    FMyStreamReader: TMyStreamReader;
    procedure AfterConstruction; override;
    procedure BeforeDestruction; override;
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

uses
  System.StrUtils
  , System.Threading
  , System.Diagnostics
  , PsAPI
  ;

// from https://stackoverflow.com/a/442278/1972425
// https://social.technet.microsoft.com/Forums/windows/en-US/73ebbbcf-ee9c-42bc-850d-fecc539cc489/windows-7-task-manager-processes-memory-private-working-set-versus-performance-memory-indicator?forum=w7itproperf
function GetProcessMemory: Cardinal;
var
  cb: Integer;
  lPMC: TProcessMemoryCounters;
begin
  // Get the used memory for the current process
  cb := SizeOf(TProcessMemoryCounters);
  lPMC.cb := cb;
  if GetProcessMemoryInfo(GetCurrentProcess(), @lPMC, cb) then
    Result:= lPMC.WorkingSetSize
  else
    RaiseLastOSError;
end;

procedure TForm1.AfterConstruction;
begin
  inherited;
  FFillControlElapsedCaption := lblFillControlElapsed.Caption;
  FVSTLinesCountCaption := lblVSTLinesCount.Caption;
  FTVLinesCountCaption := lblTVLinesCount.Caption;
  lblMemoryMax.Caption := Format(lblMemoryMax.Caption, [gauMemory.MaxValue]);
end;

procedure TForm1.BeforeDestruction;
begin
  inherited;
  FreeAndNil(FMyStreamReader);
end;

//procedure TForm1.btnReadBlocksClick(Sender: TObject);
//var
//  aFS: TFileStream;
//  aSR: TStreamReader;
//  aBuffer: TCharArray;
//  aLine: string;
//begin
//  aFS := TFileStream.Create('../../TextFile.txt', fmOpenRead);
//  try
//    aSR := TStreamReader.Create(aFS);
//    try
//      SetLength(aBuffer, 3);
//      while not aSR.EndOfStream do
//      begin
//        if aSR.ReadBlock(aBuffer, 0, Length(aBuffer)) < Length(aBuffer) then
//          MessageDlg('Reading failed! Expected to read x characters!', mtError, [mbOK], 0);
//        ListBox1.AddItem(string(aBuffer) + 'p:'+ IntToStr(aSR.BaseStream.Position), nil);
//        aSR.ReadBlock(aBuffer, 0, 2); // read carriage-return
//      end;
//    finally
//      aSR.Free;
//    end;
//  finally
//    aFS.Free;
//  end;
//end;

//procedure TForm1.btnReadLinesClick(Sender: TObject);
//var
//  aFS: TFileStream;
//  aSR: TStreamReader;
//begin
//  aFS := TFileStream.Create('../../TextFile.txt', fmOpenRead);
//  try
//    aSR := TStreamReader.Create(aFS);
//    try
//      while not aSR.EndOfStream do
//        ListBox1.AddItem(aSR.ReadLine + 'p:'+ IntToStr(aSR.BaseStream.Position), nil);
//    finally
//      aSR.Free;
//    end;
//  finally
//    aFS.Free;
//  end;
//end;

procedure TForm1.btnCreateLargeFileClick(Sender: TObject);
var
  aSW: TStreamWriter;
  s: string;
  i: Integer;
const
  cLineContent: string = 'this is line #%0:d, ';
begin
  if speFileLinesCount.Value > 100000 then
    raise Exception.Create('file size will be too large');
  aSW := TStreamWriter.Create(cLargeFile, false, TEncoding.ANSI);
  try
    for i := 1 to speFileLinesCount.Value do
    begin
      s := Format(DupeString(cLineContent, 6), [i]);
      Delete(s, 101, 1000); // calibrate the line
      aSW.WriteLine(s);
    end;
  finally
    aSW.Close;
    aSW.Free;
  end;
end;

procedure TForm1.btnFillVSTClick(Sender: TObject);
var
  lSW: TStopwatch;
begin
  if FMyStreamReader = nil then
  begin
    lSW := TStopWatch.StartNew;
    FMyStreamReader := TMyStreamReader.Create(cLargeFile);
    VirtualStringTree1.RootNodeCount := FMyStreamReader.LineCount;
    lSW.Stop;
    OutputDebugString(PChar(Format('%u', [lSW.ElapsedMilliseconds])));
  end;
//  lblFillControlElapsed.Caption := 'NA';
  lblFillControlElapsed.Caption := Format(FFillControlElapsedCaption, [lSW.ElapsedMilliseconds]);
  lblVSTLinesCount.Caption := Format(FVSTLinesCountCaption, [VirtualStringTree1.TotalCount]);
  timMemoryTrace.Enabled := true;
end;

procedure TForm1.timMemoryTraceTimer(Sender: TObject);
{$J+}
const
  lMemoryUsedCaption: string = '';
{$J-}
var
  lMemoryMB: Integer;
begin
  if lMemoryUsedCaption = '' then
    lMemoryUsedCaption := lblMemory.Caption
  else
    lblMemory.Caption := lMemoryUsedCaption;
  lMemoryMB := GetProcessMemory div 1000000;
  lblMemory.Caption := Format(lMemoryUsedCaption, [lMemoryMB]);
  gauMemory.Progress := lMemoryMB;
end;

procedure TForm1.btnClearVSTClick(Sender: TObject);
begin
  VirtualStringTree1.RootNodeCount := 0;
  FreeAndNil(FMyStreamReader);
  lblFillControlElapsed.Caption := FFillControlElapsedCaption;
  lblVSTLinesCount.Caption := FVSTLinesCountCaption;
end;

procedure TForm1.btnFillTVClick(Sender: TObject);
var
  lStreamReader: TMyStreamReader;
begin
//  if cFillControlElapsed = '' then
//    cFillControlElapsed := lblFillTVElapsed.Caption
//  else
//  lblFillControlElapsed.Caption := FFillControlElapsedCaption;
  TTask.Run(procedure
    var
      lSW: TStopwatch;
    begin
      TreeView1.Items.BeginUpdate;
      lStreamReader := TMyStreamReader.Create(cLargeFile);
      try
        // lengthy stuff here
        lSW := TStopWatch.StartNew;
        while not lStreamReader.EndOfStream do
          TreeView1.Items.AddChild(nil, lStreamReader.ReadLine());
        lSW.Stop;
        TThread.Queue(nil, procedure
          begin
            lblFillControlElapsed.Caption := Format(FFillControlElapsedCaption, [lSW.ElapsedMilliseconds]);
            lblTVLinesCount.Caption := Format(FTVLinesCountCaption, [TreeView1.Items.Count]);
          end);
      finally
        lStreamReader.Free;
        TreeView1.Items.EndUpdate;
      end;
    end);
  timMemoryTrace.Enabled := true;
end;

procedure TForm1.btnClearTVClick(Sender: TObject);
begin
//  TTask.Run(procedure
//    begin
//      TreeView1.Items.BeginUpdate;
//      TreeView1.Enabled := false;
//      try
//        TThread.Queue(nil, procedure
//          begin
            TreeView1.Items.Clear;
//          end);
//      finally
//        TreeView1.Items.EndUpdate;
//        TreeView1.Enabled := true;
//      end;
//    end);
  lblFillControlElapsed.Caption := FFillControlElapsedCaption;
  lblTVLinesCount.Caption := FTVLinesCountCaption;
end;

procedure TForm1.VirtualStringTree1GetText(Sender: TBaseVirtualTree;
  Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType;
  var CellText: string);
begin
  if Assigned(FMyStreamReader) then
    CellText := FMyStreamReader.ReadLine(Node.Index +1);
end;

//procedure TForm1.btnReadLineClick(Sender: TObject);
//var
//  aFS: TFileStream;
//  aSR: TStreamReader;
//  aLine: Integer;
//  lByteCount: Integer;
//  s: String;
//begin
//  aLine := SpinEdit1.Value;
//  aFS := TFileStream.Create('../../TextFile.txt', fmOpenRead);
//  try
//    aSR := TStreamReader.Create(aFS);
//    try
//      // evaluate the text encoding
//      s := aSR.ReadLine;
//      lByteCount := (aSR.CurrentEncoding.GetByteCount(s) +2); // + carriage-return
//      //aSR.CurrentEncoding.GetByteCount(Buffer);
//      aSR.BaseStream.Seek(((aLine -1) * lByteCount), soFromBeginning);
//      { Reset the internal data buffer, since we modified the stream position }
//      aSR.DiscardBufferedData();
//      ListBox1.AddItem(aSR.ReadLine, nil);
//    finally
//      aSR.Free;
//    end;
//  finally
//    aFS.Free;
//  end;
//end;

{ TMyStreamReader }

constructor TMyStreamReader.Create(const aFileName: TFileName);
var
  s: String;
begin
  FStreamReader := TStreamReader.Create(aFileName);
  // evaluate the text encoding
  s := FStreamReader.ReadLine;
  FLineByteCount := FStreamReader.CurrentEncoding.GetByteCount(s) +2; // + carriage-return
  FLineCount := (FStreamReader.BaseStream.Size +2) div FLineByteCount; // + carriage-return
end;

destructor TMyStreamReader.Destroy;
begin
  FStreamReader.Free;
  inherited;
end;

function TMyStreamReader.GetEndOfStream: boolean;
begin
  result := FStreamReader.EndOfStream;
end;

function TMyStreamReader.GetLineCount: Integer;
begin
  result := FLineCount;
end;

function TMyStreamReader.ReadLine(): string;
begin
  result := FStreamReader.ReadLine;
end;

function TMyStreamReader.ReadLine(const aLine: integer): string;
begin
  if FStreamReader.BaseStream.Seek(((aLine -1) * FLineByteCount), soFromBeginning) < FStreamReader.BaseStream.Size then
  begin
    FStreamReader.DiscardBufferedData();
    result := FStreamReader.ReadLine;
  end;
end;

end.
